# ---------------------------------------------------------------------------
# MTDA Scripts for IPC227E
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import time


def ipc227e_enter_bios():
    # system may take a while to come up if SecureBoot settings were reset
    # send a storm of esc key presses for 50 seconds to make sure it enters
    # the BIOS.
    for i in range(5):
        start = time.time()
        while (time.time() - start) < 10:
            mtda.keyboard.esc(5)
        sleep(0.1)


def ipc227e_enter_secure_boot_option():
    mtda.keyboard.right()
    mtda.keyboard.down()
    mtda.keyboard.enter()
    sleep(10)


def ipc227e_disable_secure_boot():
    mtda.keyboard.down(3)
    mtda.keyboard.up(2)
    mtda.keyboard.enter()
    mtda.keyboard.up()
    sleep(1)
    mtda.keyboard.enter()


def ipc227e_enable_secure_boot():
    mtda.keyboard.down(3)
    mtda.keyboard.up(2)
    mtda.keyboard.enter()
    mtda.keyboard.down()
    sleep(1)
    mtda.keyboard.enter()


def ipc227e_erase_secure_boot_settings():
    mtda.keyboard.down(3)
    mtda.keyboard.up()
    mtda.keyboard.enter()
    mtda.keyboard.down()
    sleep(1)
    mtda.keyboard.enter()


def ipc227e_restore_secure_boot_factory_settings():
    mtda.keyboard.down(3)
    mtda.keyboard.enter()
    mtda.keyboard.down()
    sleep(1)
    mtda.keyboard.enter()


def ipc227e_apply_secure_boot_settings():
    mtda.keyboard.f10()
    sleep(1)
    mtda.keyboard.enter()
    sleep(5)


def ipc227e_reset_tpm():
    ipc227e_enter_bios()
    ipc227e_enter_secure_boot_option()
    ipc227e_disable_secure_boot()
    ipc227e_erase_secure_boot_settings()
    ipc227e_apply_secure_boot_settings()
    return 0


def ipc227e_reset_tpm_factory():
    ipc227e_enter_bios()
    ipc227e_enter_secure_boot_option()
    ipc227e_disable_secure_boot()
    ipc227e_restore_secure_boot_factory_settings()
    ipc227e_apply_secure_boot_settings()
    return 0


def ipc227e_enable_secureboot():
    ipc227e_enter_bios()
    ipc227e_enter_secure_boot_option()
    ipc227e_enable_secure_boot()
    ipc227e_apply_secure_boot_settings()
    return 0


def ipc227e_disable_secureboot():
    ipc227e_enter_bios()
    ipc227e_enter_secure_boot_option()
    ipc227e_disable_secure_boot()
    ipc227e_apply_secure_boot_settings()
    return 0


def ipc227e_boot_from_usb():
    ipc227e_enter_bios()
    mtda.keyboard.right()
    mtda.keyboard.enter()
    sleep(0.5)
    mtda.keyboard.down(10)
    mtda.keyboard.enter()
    mtda.keyboard.idle()
    return 0


def ipc227e_power_on():
    scripts.check_reset_tpm() or \
        scripts.check_reset_tpm_factory() or \
        scripts.check_disable_secureboot() or \
        scripts.check_enable_secureboot() or \
        scripts.check_boot_from_usb()



# ---------------------------------------------------------------------------
# mjpg_streamer driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import atexit
import os
import psutil
import signal
import socket
import threading
import time

# Local imports
from mtda.video.controller import VideoController


class MJPGStreamerVideoController(VideoController):

    def __init__(self, mtda):
        self.dev = "/dev/video0"
        self.executable = "mjpg_streamer"
        self.input = "-i 'input_uvc.so -d {0} -r {1}'"
        self.lock = threading.Lock()
        self.mtda = mtda
        self.output = "-o 'output_http.so -p {0} -w {1}'"
        self.pid = None
        self.port = 8080
        self.resolution = "1280x780"
        self.www = "/usr/share/mjpg-streamer/www"

    def configure(self, conf):
        self.mtda.debug(3, "video.mjpg_streamer.configure()")

        if 'device' in conf:
            self.dev = conf['device']
            self.mtda.debug(4, "video.mjpg_streamer.configure(): "
                               "will use %s" % str(self.dev))
        if 'executable' in conf:
            self.executable = conf['executable']
        if 'port' in conf:
            self.port = conf['port']
        if 'resolution' in conf:
            self.resolution = conf['resolution']
        if 'www' in conf:
            self.www = conf['www']

    def configure_systemd(self, dir):
        device = os.path.basename(self.dev)
        dropin = os.path.join(dir, 'auto-dep-video.conf')
        with open(dropin, 'w') as f:
            f.write('[Unit]\n')
            f.write('Wants=dev-{}.device\n'.format(device))
            f.write('After=dev-{}.device\n'.format(device))

    @property
    def format(self):
        return "MJPG"

    def probe(self):
        self.mtda.debug(3, "video.mjpg_streamer.probe()")

        if self.executable is None:
            raise ValueError("mjpg_streamer executable not specified!")

        result = True
        if os.system("%s --version" % self.executable) != 0:
            raise ValueError("could not execute %s!" % self.executable)
        if os.path.exists(self.dev) is False:
            self.mtda.debug(1, "video.mjpg_streamer.probe(): "
                               "video device (%s) not found!"
                               % self.dev)
            result = False

        self.mtda.debug(3, "video.mjpg_streamer.probe(): %s" % str(result))
        return result

    def getpid(self):
        name = os.path.basename(self.executable)
        pid = None

        for proc in psutil.process_iter():
            if name in proc.name():
                pid = proc.pid
                break

        return pid

    def start(self):
        self.mtda.debug(3, "video.mjpg_streamer.start()")

        if self.pid is not None:
            return True

        atexit.register(self.stop)

        options = self.input.format(self.dev, self.resolution)
        options += " "
        options += self.output.format(self.port, self.www)
        options += " -b"

        result = os.system("%s %s" % (self.executable, options))
        if result == 0:
            self.pid = self.getpid()
            self.mtda.debug(2, "video.mjpg_streamer.start(): "
                               "mjpg_streamer process started "
                               "[{0}]".format(self.pid))
            return True
        else:
            self.mtda.debug(1, "video.mjpg_streamer.start(): "
                               "mjpg_streamer process failed "
                               "({0})".format(result))
        return False

    def kill(self, name, pid, timeout=3):
        tries = timeout
        if psutil.pid_exists(pid):
            self.mtda.debug(2, "terminating {0} "
                               "[{1}] using SIGTERM".format(name, pid))
            os.kill(pid, signal.SIGTERM)
        while tries > 0 and psutil.pid_exists(pid):
            time.sleep(1)
            tries = tries - 1
        if psutil.pid_exists(pid):
            self.mtda.debug(2, "terminating {0} "
                               "[{1}] using SIGKILL".format(name, pid))
            os.kill(pid, signal.SIGKILL)
        return psutil.pid_exists(pid)

    def stop(self):
        self.mtda.debug(3, "video.mjpg_streamer.stop()")

        self.lock.acquire()
        result = True

        if self.pid is not None:
            name = os.path.basename(self.executable)
            result = self.kill(name, self.pid)
            if result:
                self.pid = None

        self.lock.release()
        self.mtda.debug(3, "video.mjpg_streamer.stop(): %s" % str(result))
        return result

    def url(self, host="", opts=None):
        self.mtda.debug(3, "video.mjpg_streamer.url(host='%s')" % str(host))

        if host is None or host == "":
            host = socket.getfqdn()
            self.mtda.debug(3, "video.mjpg_streamer.url: "
                               "using host='%s'" % str(host))
        result = "http://{0}:{1}/?action=stream".format(host, self.port)

        self.mtda.debug(3, "video.mjpg_streamer.url(): %s" % str(result))
        return result


def instantiate(mtda):
    return MJPGStreamerVideoController(mtda)



# ---------------------------------------------------------------------------
# Video interface for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import abc


class VideoController(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def configure(self, conf):
        """ Configure this video controller from the provided
            configuration"""
        return

    @property
    def format(self):
        return None

    @abc.abstractmethod
    def probe(self):
        """ Check presence of the video controller"""
        return

    @abc.abstractmethod
    def start(self):
        """ Start video capture """
        return False

    @abc.abstractmethod
    def stop(self):
        """ Stop video capture """
        return False

    @abc.abstractmethod
    def url(self, host="", opts=None):
        """ URL for the video stream """
        return None



# ---------------------------------------------------------------------------
# usbrelay power driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import os
import re
import subprocess

# Local imports
from mtda.power.controller import PowerController


class UsbRelayPowerController(PowerController):

    def __init__(self, mtda):
        self.dev = None
        self.exe = "/usr/bin/usbrelay"
        self.mtda = mtda
        self.lines = []

    def configure(self, conf):
        if 'lines' in conf:
            self.lines = conf['lines'].split(',')

    def probe(self):
        if self.lines is None:
            raise ValueError("usbrelay: 'lines' not configured!")

        if os.path.exists(self.exe) is False:
            raise ValueError("{0} was not found!".format(self.exe))

        # Make sure all configured lines were detected
        statuses = self._get_lines()
        for line in self.lines:
            if line not in statuses:
                raise ValueError("usbrelay: {0} not detected!".format(line))

    def command(self, args):
        return False

    def on(self):
        if self._set_lines("1") is False:
            return False
        return self.status() == self.POWER_ON

    def off(self):
        if self._set_lines("0") is False:
            return False
        return self.status() == self.POWER_OFF

    def status(self):
        statuses = self._get_lines()
        first = True
        result = self.POWER_UNSURE
        for line in self.lines:
            value = statuses[line]
            if value == '1':
                value = self.POWER_ON
            else:
                value = self.POWER_OFF
            self.mtda.debug(3, "power.usbrelay.status: "
                               "line {0} is {1}".format(line, value))
            if first is True:
                first = False
                result = value
            elif value != result:
                result = self.POWER_UNSURE
        return result

    def _get_lines(self):
        result = {}
        try:
            lines = subprocess.check_output(
                    [self.exe]).decode("utf-8").splitlines()
            for line in lines:
                m = re.match(r"(\w+)=([01])", line)
                if m is not None:
                    name = m[1]
                    value = m[2]
                    result[name] = value
        except subprocess.CalledProcessError:
            result = None

        return result

    def _set_lines(self, value):
        try:
            for line in self.lines:
                subprocess.check_call(
                        [self.exe, "{0}={1}".format(line, value)])
        except subprocess.CalledProcessError:
            return False
        return True


def instantiate(mtda):
    return UsbRelayPowerController(mtda)



# ---------------------------------------------------------------------------
# Shell command power driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (c) Siemens AG, 2022
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import subprocess
import threading

# Local imports
from mtda.power.controller import PowerController


class ShellCmdPowerController(PowerController):

    def __init__(self, mtda):
        self.ev = threading.Event()
        self.on_cmd = None
        self.off_cmd = None
        self.check_on = None
        self.mtda = mtda

    def configure(self, conf):
        if 'on-cmd' in conf:
            self.on_cmd = conf['on-cmd']
        if 'off-cmd' in conf:
            self.off_cmd = conf['off-cmd']
        if 'check-on' in conf:
            self.check_on = conf['check-on']

    def probe(self):
        if self.on_cmd is None:
            raise ValueError("on-cmd not specified")
        self.mtda.debug(3, "on-cmd: {}".format(self.on_cmd))
        if self.off_cmd is None:
            raise ValueError("off-cmd not specified")
        self.mtda.debug(3, "off-cmd: {}".format(self.off_cmd))
        if self.check_on is None:
            raise ValueError("check-on not specified")
        self.mtda.debug(3, "check_on: {}".format(self.check_on))

    def command(self, args):
        return False

    def on(self):
        proc = subprocess.run(self.on_cmd, shell=True, capture_output=True)
        if proc.returncode != 0:
            return False
        self.ev.set()
        return True

    def off(self):
        proc = subprocess.run(self.off_cmd, shell=True, capture_output=True)
        if proc.returncode != 0:
            return False
        self.ev.clear()
        return True

    def status(self):
        proc = subprocess.run(self.check_on, shell=True, capture_output=True)
        self.mtda.debug(3, "check_on: {}".format(proc.returncode))
        if proc.returncode == 0:
            return self.POWER_ON
        elif proc.returncode == 1:
            return self.POWER_OFF
        else:
            self.POWER_UNSURE

    def toggle(self):
        s = self.status()
        if s == self.POWER_OFF:
            self.on()
        else:
            self.off()
        return self.status()

    def wait(self):
        while self.status() != self.POWER_ON:
            self.ev.wait()


def instantiate(mtda):
    return ShellCmdPowerController(mtda)



# ---------------------------------------------------------------------------
# Support for USB Composite
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import atexit
import os
import threading


class Composite:

    mtda = None
    lock = threading.Lock()
    vendor_id = "0x1d6b"  # Linux Foundation
    product_id = "0x0104"  # Multifunction Composite Gadget
    serialnumber = "mtda-2020"
    manufacturer = "Siemens Digital Industries Software"
    product = "USB functions for Multi-Tenant Device Access"
    lang = "0x409"  # english
    _storage = None
    _installed = False

    path = "/sys/kernel/config/usb_gadget/" + product.lower().replace(" ", "_")

    def debug(level, msg):
        if Composite.mtda is not None:
            Composite.mtda.debug(level, msg)

    def configure(what, conf):
        Composite.debug(3, "composite.configure('{}')".format(what))

        with Composite.lock:
            result = Composite._configure(what, conf)

        Composite.debug(3, "composite.configure(): {}".format(result))
        return result

    def _configure(what, conf):
        result = True
        if what == 'storage' and 'file' in conf:
            Composite._storage = conf['file']
        if what in Composite.functions:
            Composite.functions[what]['configured'] = True
            Composite.functions[what]['enabled'] = True
            Composite.debug(2, "composite.configure(): "
                               "{} configured & enabled".format(what))
        else:
            Composite.debug(1, "composite.configure(): "
                               "not supported")
            result = False

        return result

    def _enable():
        usbdrv = os.listdir("/sys/class/udc")[0]
        udc = os.path.join(Composite.path, "UDC")
        result = write(udc, usbdrv)
        return result > 0

    def _disable():
        udc = os.path.join(Composite.path, "UDC")
        if os.path.exists(udc):
            write(udc, "")
        return True

    def install():
        Composite.debug(3, "composite.install()")

        with Composite.lock:
            result = Composite._install()

        Composite.debug(3, "composite.install(): {}".format(result))
        return result

    def _install():
        if Composite._installed is True:
            return True

        Composite._remove()

        enabled = False
        for function in Composite.functions.values():
            if function['enabled'] is True:
                enabled = True
                break

        if enabled is False:
            # Nothing to install
            Composite.debug(2, "composite.install(): "
                               "no functions were enabled")
            return True

        atexit.register(Composite.remove)
        path = Composite.path
        create_dirs(path)

        write(path + "/idVendor",  Composite.vendor_id)
        write(path + "/idProduct", Composite.product_id)
        write(path + "/bcdDevice", "0x0100")
        write(path + "/bcdUSB", "0x0200")

        strings = path + "/strings/" + Composite.lang + "/"
        write(strings + "serialnumber", Composite.serialnumber)
        write(strings + "manufacturer", Composite.manufacturer)
        write(strings + "product", Composite.product)

        strings = path + "/configs/c.1/strings/" + Composite.lang + "/"
        write(strings + "configuration", "Config 1: ECM network")
        write(path + "/configs/c.1/MaxPower", "250")

        Composite._create_functions()

        if Composite.functions['storage']['enabled'] is True:
            lun = path + "/functions/mass_storage.usb0/lun.0/"
            file = Composite._storage
            write(lun + "cdrom", "0")
            write(lun + "ro", "0")
            write(lun + "nofua", "0")
            write(lun + "file", file)
            Composite.debug(2, "composite.install(): "
                               "storage device/file: {}".format(file))
        Composite._installed = Composite._enable()
        return Composite._installed

    def remove():
        Composite.debug(3, "composite.remove()")

        with Composite.lock:
            result = Composite._remove()

        Composite.debug(3, "composite.remove(): {}".format(result))
        return result

    def _remove():
        lang = Composite.lang
        path = Composite.path

        if os.path.exists(path) is False:
            return

        Composite._disable()

        functions = os.listdir(os.path.join(path, "functions"))
        for function in functions:
            os.unlink(os.path.join(path, "configs", "c.1", function))
        os.rmdir(os.path.join(path, "configs", "c.1", "strings", lang))
        os.rmdir(os.path.join(path, "configs", "c.1"))
        for function in functions:
            os.rmdir(os.path.join(path, "functions", function))
        os.rmdir(os.path.join(path, "strings", lang))
        os.rmdir(path)

        Composite._installed = False

    def _create_functions():
        for function in Composite.functions.values():
            name = function['name']
            if function['configured'] is False:
                continue
            if function['enabled'] is False:
                continue
            Composite.debug(2, "composite._create_functions: "
                               "registering {}".format(name))
            path = Composite.path + "/functions/" + name
            if not os.path.exists(path):
                os.makedirs(path)
            if 'protocol' in function:
                write(path + "/protocol", function['protocol'])
            if 'subclass' in function:
                write(path + "/subclass", function['subclass'])
            if 'report_length' in function:
                write(path + "/report_length", function['report_length'])
            if 'report_desc' in function:
                write(path + "/report_desc", function['report_desc'], "wb")
            config = Composite.path + "/configs/c.1/" + name
            if not os.path.exists(config):
                os.symlink(path, config, True)

    def storage_toggle(enabled):
        with Composite.lock:
            if Composite.functions['storage']['configured'] is True:
                Composite.functions['storage']['enabled'] = enabled

    hid_function = {
        "name": "hid.usb0",
        "configured": False,
        "enabled": False,
        "protocol": "1",
        "subclass": "1",
        "report_length": "8",
        "report_desc": [
            0x05, 0x01,  # USAGE_PAGE (Generic Desktop)
            0x09, 0x06,  # USAGE (Keyboard)
            0xa1, 0x01,  # COLLECTION (Application)
            0x05, 0x07,  # USAGE_PAGE (Keyboard)
            0x19, 0xe0,  # USAGE_MINIMUM (Keyboard LeftControl)
            0x29, 0xe7,  # USAGE_MAXIMUM (Keyboard Right GUI)
            0x15, 0x00,  # LOGICAL_MINIMUM (0)
            0x25, 0x01,  # LOGICAL_MAXIMUM (1)
            0x75, 0x01,  # REPORT_SIZE (1)
            0x95, 0x08,  # REPORT_COUNT (8)
            0x81, 0x02,  # INPUT (Data,Var,Abs)
            0x95, 0x01,  # REPORT_COUNT (1)
            0x75, 0x08,  # REPORT_SIZE (8)
            0x81, 0x01,  # INPUT (Cnst,Var,Abs) // 0x03
            0x95, 0x05,
            0x75, 0x01,
            0x05, 0x08,
            0x19, 0x01,
            0x29, 0x05,
            0x91, 0x02,
            0x95, 0x01,
            0x75, 0x03,
            0x91, 0x01,  # 0x03
            0x95, 0x06,  # REPORT_COUNT (6)
            0x75, 0x08,  # REPORT_SIZE (8)
            0x15, 0x00,  # LOGICAL_MINIMUM (0)
            0x25, 0x65,  # LOGICAL_MAXIMUM (101)
            0x05, 0x07,  # USAGE_PAGE (Keyboard)
            0x19, 0x00,  # USAGE_MINIMUM (Reserved (no event indicated))
            0x29, 0x65,  # USAGE_MAXIMUM (Keyboard Application)
            0x81, 0x00,  # INPUT (Data,Ary,Abs)
            0xc0
        ]
    }

    console_function = {
        "name": "acm.GS0",
        "configured": False,
        "enabled": False
    }

    monitor_function = {
        "name": "acm.GS1",
        "configured": False,
        "enabled": False
    }

    ms_function = {
        "name": "mass_storage.usb0",
        "configured": False,
        "enabled": False
    }

    functions = {
        "console": console_function,
        "monitor": monitor_function,
        "keyboard": hid_function,
        "storage": ms_function
    }


def write(filename, content, mode="w"):
    with open(filename, mode) as fp:
        if type(content) == str:
            return fp.write(content)
        if type(content) == list:
            return fp.write(bytearray(content))


def create_dirs(path):
    if not os.path.exists(path):
        os.makedirs(path)

    if not os.path.exists(path + "/strings/" + Composite.lang):
        os.makedirs(path + "/strings/" + Composite.lang)

    if not os.path.exists(path + "/configs/c.1/strings/" + Composite.lang):
        os.makedirs(path + "/configs/c.1/strings/" + Composite.lang)



# ---------------------------------------------------------------------------
# Console input for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import atexit
import codecs
import fcntl
import termios
import sys
import tty


class ConsoleInput:

    def __init__(self):
        if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
            sys.stdin.reconfigure(encoding='utf-8', errors='ignore')
        else:
            sys.stdin = codecs.getreader('utf-8')(sys.stdin.detach())
        if sys.stdin.isatty():
            self.fd = sys.stdin.fileno()
            self.old = termios.tcgetattr(self.fd)
            atexit.register(self.cleanup)

    def start(self):
        if sys.stdin.isatty():
            new = termios.tcgetattr(self.fd)
            new[3] = new[3] & ~termios.ICANON & ~termios.ECHO & ~termios.ISIG
            new[6][termios.VMIN] = 1
            new[6][termios.VTIME] = 0
            termios.tcsetattr(self.fd, termios.TCSANOW, new)
            tty.setraw(sys.stdin)

    def getkey(self):
        c = sys.stdin.read(1)
        if c == chr(0x7f):
            c = chr(8)  # map the BS key (which yields DEL) to backspace
        return c

    def cancel(self):
        if sys.stdin.isatty():
            fcntl.ioctl(self.fd, termios.TIOCSTI, b'\0')

    def cleanup(self):
        if sys.stdin.isatty():
            termios.tcsetattr(self.fd, termios.TCSADRAIN, self.old)



# ---------------------------------------------------------------------------
# docker console driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import array
import fcntl
import os
import select
import termios

# Local imports
from mtda.console.interface import ConsoleInterface


class DockerConsole(ConsoleInterface):

    def __init__(self, mtda):
        self.mtda = mtda
        self.docker = mtda.power_controller
        self._fd = None
        self._opened = False
        self._socket = None

    """ Configure this console from the provided configuration"""
    def configure(self, conf, role='console'):
        self.mtda.debug(3, "console.docker.configure()")

    def probe(self):
        self.mtda.debug(3, "console.docker.probe()")
        result = self.docker.variant == "docker"
        self.mtda.debug(3, "console.docker.probe(): {}".format(result))
        return result

    def open(self):
        self.mtda.debug(3, "console.docker.open()")

        result = self._opened
        if self._opened is False:
            self._socket = self.docker.socket()
            if self._socket is not None:
                self._fd = self._socket.fileno()
                result = True
            else:
                self.mtda.debug(1, "console.docker.open(): "
                                "could not attach to container socket")
        else:
            self.mtda.debug(4, "console.docker.open(): already opened")

        self._opened = result
        self.mtda.debug(3, "console.docker.open(): {}".format(result))
        return result

    def close(self):
        self.mtda.debug(3, "console.docker.close()")

        result = True
        if self._opened is True:
            self._socket.close()
            self._fd = None
            self._socket = None
            self._opened = False

        self.mtda.debug(3, "console.docker.close(): {}".format(result))
        return result

    """ Return number of pending bytes to read"""
    def pending(self):
        self.mtda.debug(3, "console.docker.pending()")

        result = 0
        if self._opened is True:
            avail = array.array('l', [0])
            fcntl.ioctl(self._fd, termios.FIONREAD, avail, 1)
            result = avail[0]

        self.mtda.debug(3, "console.docker.pending(): {}".format(result))
        return result

    """ Read bytes from the console"""
    def read(self, n=1):
        self.mtda.debug(3, "console.docker.read({})".format(n))

        result = None
        if self._opened is True:
            inputs = [self._socket]
            outputs = []
            readable, writable, error = select.select(inputs, outputs, inputs)
            if len(readable) > 0 and self._fd is not None:
                result = os.read(self._fd, n)
        else:
            self.mtda.debug(1, "console.docker.read(): not opened!")

        if result is None:
            result = b''

        self.mtda.debug(3, "console.docker.read(): {}".format(result))
        return result

    """ Write to the console"""
    def write(self, data):
        self.mtda.debug(3, "console.docker.write(data={})".format(data))

        result = None
        if self._opened is True:
            result = os.write(self._fd, data)

        self.mtda.debug(3, "console.docker.write(): {}".format(result))
        return result


def instantiate(mtda):
    return DockerConsole(mtda)



# ---------------------------------------------------------------------------
# USB Function console driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# Local imports
from mtda.console.serial import SerialConsole
from mtda.support.usb import Composite


class UsbFunctionConsole(SerialConsole):

    def __init__(self, mtda):
        super().__init__(mtda)
        self.hotplug = True
        self.port = None
        self.rate = 9600
        Composite.mtda = mtda

    def configure(self, conf, role='console'):
        self.mtda.debug(3, "console.usbf.configure()")

        super().configure(conf)
        if self.port is None:
            self.port = "/dev/ttyGS0" if role == "console" else "/dev/ttyGS1"
        result = Composite.configure(role, conf)

        self.mtda.debug(3, "console.usbf.configure(): {}".format(result))
        return result

    def configure_systemd(self, dir):
        return None


def instantiate(mtda):
    return UsbFunctionConsole(mtda)



# ---------------------------------------------------------------------------
# GPIO usb driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import gpiod
from operator import itemgetter

# Local imports
from mtda.usb.switch import UsbSwitch


class GpioUsbSwitch(UsbSwitch):

    def __init__(self, mtda):
        self.dev = None
        self.pin = 0
        self.enable = 1
        self.disable = 0
        self.mtda = mtda
        self.lines = []
        self.gpiopair = []

    def configure(self, conf):
        """ Configure this USB switch from the provided configuration"""
        self.mtda.debug(3, "usb.gpio.configure()")
        if 'pin' in conf:
            self.pin = int(conf['pin'], 10)
        if 'enable' in conf:
            if conf['enable'] == 'high':
                self.enable = 1
                self.disable = 0
            elif conf['enable'] == 'low':
                self.enable = 0
                self.disable = 1
            else:
                raise ValueError("'enable' shall be either 'high' or 'low'!")
        if 'gpio' in conf:
            for gpio in conf['gpio'].split(','):
                self.gpiopair.append(itemgetter(0, 1)(gpio.split('@')))

        result = True
        self.mtda.debug(3, "usb.gpio.configure(): %s" % str(result))
        return result

    def probe(self):
        self.mtda.debug(3, "usb.gpio.probe()")
        if not self.gpiopair:
            raise ValueError("GPIO chip(s) and pin(s) pair not configured")

        for chip, pin in self.gpiopair:
            pin = int(pin)
            chip = gpiod.Chip(chip, gpiod.Chip.OPEN_BY_NAME)
            self.mtda.debug(3, "this is chip {} and pin is {} "
                               "power.gpio.configure(): ".format(chip, pin))
            self.lines.append(chip.get_line(pin))

        for line in self.lines:
            try:
                if line.is_used() is False:
                    self.mtda.debug(3, "power.gpiochip{}@pin{} is free for "
                                       "use" .format(chip, pin))
                    line.request(consumer='mtda', type=gpiod.LINE_REQ_DIR_OUT)
                else:
                    raise ValueError("gpiochip{}@pin{} is in use by other "
                                     "service" .format(chip, pin))
            except OSError:
                self.mtda.debug(3, "line {} is not configured correctly"
                                   .format(line))

        result = True
        self.mtda.debug(3, "usb.gpio.probe(): {}" .format(result))
        return result

    def on(self):
        """ Power on the target USB port"""
        self.mtda.debug(3, "usb.gpio.on()")
        for line in self.lines:
            line.set_value(self.enable)
        result = self.status() == self.POWERED_ON
        self.mtda.debug(3, "usb.gpio.on(): {}" .format(result))
        return result

    def off(self):
        """ Power off the target USB port"""
        self.mtda.debug(3, "usb.gpio.off()")
        for line in self.lines:
            line.set_value(self.disable)
        result = self.status() == self.POWERED_OFF
        self.mtda.debug(3, "usb.gpio.off(): {}" .format(result))
        return result

    def status(self):
        """ Determine the current power state of the USB port"""
        self.mtda.debug(3, "usb.gpio.status()")
        value = self.line.get_value()
        for line in self.lines:
            value = line.get_value()
            if value == self.enable:
                result = self.POWERED_ON
            else:
                result = self.POWERED_OFF

            self.mtda.debug(3, "usb.gpio.status():"
                               "line {} is {}" .format(line, value))
        return result

    def toggle(self):
        self.mtda.debug(3, "usb.gpio.toggle()")
        s = self.status()
        if s == self.POWERED_ON:
            self.off()
            result = self.POWERED_OFF
        else:
            self.on()
            result = self.POWERED_ON

        self.mtda.debug(3, "usb.gpio.toggle(): %s" % str(result))
        return result


def instantiate(mtda):
    return GpioUsbSwitch(mtda)



# ---------------------------------------------------------------------------
# QEMU power driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import atexit
import os
import pathlib
import psutil
import re
import signal
import tempfile
import threading
import time
import multiprocessing

# Local imports
from mtda.power.controller import PowerController


class QemuController(PowerController):

    def __init__(self, mtda):
        self.dev = None
        self.bios = None
        self.cpu = None
        self.smp = None
        self.drives = []
        self.executable = "kvm"
        self.hostname = "mtda-kvm"
        self.lock = threading.Lock()
        self.machine = None
        self.memory = 512
        self.mtda = mtda
        self.novnc = "/usr/share/novnc"
        self.pflash_ro = None
        self.pflash_rw = None
        self.pidOfQemu = None
        self.pidOfSwTpm = None
        self.pidOfWebsockify = None
        self.swtpm = "/usr/bin/swtpm"
        self.watchdog = None
        self.websockify = "/usr/bin/websockify"

    def configure(self, conf):
        self.mtda.debug(3, "power.qemu.configure()")

        if 'bios' in conf:
            self.bios = conf['bios']
        if 'cpu' in conf:
            self.cpu = conf['cpu']
        if 'smp' in conf:
            self.smp = int(conf['smp'])
        if 'executable' in conf:
            self.executable = conf['executable']
        if 'hostname' in conf:
            self.hostname = conf['hostname']
        if 'machine' in conf:
            self.machine = conf['machine']
        if 'memory' in conf:
            self.memory = int(conf['memory'])
        if 'pflash_ro' in conf:
            self.pflash_ro = os.path.realpath(conf['pflash_ro'])
        if 'pflash_rw' in conf:
            self.pflash_rw = os.path.realpath(conf['pflash_rw'])
        if 'storage' in conf and 'storage.0' not in conf:
            conf['storage.0'] = conf['storage']
        if 'storage.size' in conf and 'storage.0.size' not in conf:
            conf['storage.0.size'] = conf['storage.size']
        if 'swtpm' in conf:
            self.swtpm = os.path.realpath(conf['swtpm'])
        elif os.path.exists(self.swtpm) is False:
            self.swtpm = None
        if 'watchdog' in conf:
            self.watchdog = conf['watchdog']
        n = 0
        while True:
            key = 'storage.{}'.format(n)
            sizekey = 'storage.{}.size'.format(n)
            if key in conf:
                path = os.path.realpath(conf[key])
                size = int(conf[sizekey]) if sizekey in conf else 16
                self.drives.append((path, size))
                n = n + 1
            else:
                break

    def probe(self):
        self.mtda.debug(3, "power.qemu.probe()")

        if self.executable is None:
            raise ValueError("qemu executable not specified!")
        result = os.system("%s --version" % self.executable)
        if result != 0:
            raise ValueError("could not execute %s!" % self.executable)
        if self.swtpm is not None and os.path.exists(self.swtpm) is False:
            raise ValueError("swtpm (%s) could not be found!" % self.swtpm)

    def getpid(self, pidfile, timeout=30):
        result = 0
        while timeout > 0:
            with open(pidfile, "r") as f:
                data = f.read()
                if data:
                    result = int(data)
                    break
            time.sleep(1)
            timeout = timeout - 1
        return result

    def getproc(self, cmd):
        for proc in psutil.process_iter():
            try:
                proccmd = " ".join(proc.cmdline())
                if proccmd.endswith(cmd):
                    return proc.pid
            except (psutil.NoSuchProcess,
                    psutil.AccessDenied,
                    psutil.ZombieProcess):
                pass
        return None

    def start(self):
        self.mtda.debug(3, "power.qemu.start()")

        if self.pidOfQemu is not None:
            return True
        if os.path.exists("/tmp/qemu-mtda.in"):
            os.unlink("/tmp/qemu-mtda.in")
        if os.path.exists("/tmp/qemu-mtda.out"):
            os.unlink("/tmp/qemu-mtda.out")
        if os.path.exists("/tmp/qemu-serial.in"):
            os.unlink("/tmp/qemu-serial.in")
        if os.path.exists("/tmp/qemu-serial.out"):
            os.unlink("/tmp/qemu-serial.out")
        os.mkfifo("/tmp/qemu-mtda.in")
        os.mkfifo("/tmp/qemu-mtda.out")
        os.mkfifo("/tmp/qemu-serial.in")
        os.mkfifo("/tmp/qemu-serial.out")

        atexit.register(self.stop)

        # base options
        options = "-daemonize -S -m %d" % self.memory
        options += " -chardev pipe,id=monitor,path=/tmp/qemu-mtda"
        options += " -monitor chardev:monitor"
        options += " -serial pipe:/tmp/qemu-serial"
        options += " -device e1000,netdev=net0"
        options += " -netdev user,id=net0,"
        options += "hostfwd=tcp::2222-:22,hostname={0}".format(self.hostname)
        options += " -usb"
        options += " -vnc :0,websocket"

        # extra options
        if self.bios is not None:
            options += " -bios %s" % self.bios
        if self.cpu is not None:
            options += " -cpu %s" % self.cpu
        if self.smp is not None:
            if self.smp == 0:
                options += " -smp %s" % multiprocessing.cpu_count()
            else:
                options += " -smp %s" % self.smp
        if self.machine is not None:
            options += " -machine %s" % self.machine
        if self.pflash_ro is not None:
            if pathlib.Path(self.pflash_ro).is_file():
                if os.access(self.pflash_ro, os.R_OK):
                    options += " -drive if=pflash,format=raw,"
                    options += "readonly,file=%s" % self.pflash_ro
                else:
                    raise ValueError("Read-only pflash file (%s) "
                                     "cannot be read." % self.pflash_ro)
            else:
                raise ValueError("Read-only pflash file (%s) does not "
                                 "exist or is not a file." % self.pflash_ro)
        if self.pflash_rw is not None:
            try:
                options += " -drive if=pflash,format=raw,"
                options += "file=%s" % self.pflash_rw
                if pathlib.Path(self.pflash_rw).is_file():
                    if not os.access(self.pflash_rw, os.W_OK):
                        raise ValueError("Writeable pflash file (%s) has no "
                                         "write permission." % self.pflash_rw)
                else:
                    # This is probably recoverable, we'll create an empty file
                    # and trust the try/except to save us if the specified
                    # location isn't usable for some reason (e.g. the location
                    # is write protected or the specified file already exists
                    # as a directory or something).
                    #
                    # The existing OVMF fd file is ~2MB so we'll create our
                    # writeable copy at the same size.
                    sparse = pathlib.Path(self.pflash_rw)
                    sparse.touch()
                    os.truncate(str(sparse), 2*1024*1024)
            except Exception as e:
                raise ValueError("Writeable pflash file (%s) does not exist "
                                 "or is not a file and cannot be created: "
                                 "%s" % (self.pflash_rw, e))
        if len(self.drives) > 0:
            for drv, size in self.drives:
                options += " -drive file={},media=disk,format=raw".format(drv)
                if os.path.exists(drv) is False:
                    sparse = pathlib.Path(drv)
                    sparse.touch()
                    os.truncate(str(sparse), size*1024*1024*1024)
        if self.watchdog is not None:
            options += " -watchdog %s" % self.watchdog

        # swtpm options
        if self.swtpm is not None:
            with tempfile.NamedTemporaryFile() as pidfile:
                os.makedirs("/tmp/qemu-swtpm", exist_ok=True)
                result = os.system(
                      self.swtpm
                      + " socket -d"
                      + " --tpmstate dir=/tmp/qemu-swtpm"
                      + " --ctrl type=unixio,path=/tmp/qemu-swtpm/sock"
                      + " --pid file=%s --tpm2" % pidfile.name)
                if result == 0:
                    self.pidOfSwTpm = self.getpid(pidfile.name)
                    self.mtda.debug(2, "power.qemu.start(): "
                                       "swtpm process started "
                                       "[{0}]".format(self.pidOfSwTpm))
                else:
                    self.mtda.debug(1, "power.qemu.start(): "
                                       "swtpm process failed "
                                       "({0})".format(result))
                    return False

                options += " -chardev socket,id=chrtpm,"
                options += "path=/tmp/qemu-swtpm/sock"
                options += " -tpmdev emulator,id=tpm0,chardev=chrtpm"
                options += " -device tpm-tis,tpmdev=tpm0"

        # Create a WebSocket proxy to QEMU's VNC service to support noVNC
        # when our web service is enabled and have websockify installed
        if self.mtda.www is not None and os.path.exists(self.websockify):
            # bind on the same address as our web service
            host = self.mtda.www.host
            cmd = self.websockify + " -D {}:5901 {}:5900".format(host, host)
            result = os.system(cmd)
            if result == 0:
                self.pidOfWebsockify = self.getproc(cmd)
                self.mtda.debug(2, "power.qemu.start(): "
                                   "websockify process started "
                                   "[{0}]".format(self.pidOfWebsockify))

        with tempfile.NamedTemporaryFile() as pidfile:
            options += " -pidfile {0}".format(pidfile.name)
            result = os.system("%s %s" % (self.executable, options))
            if result == 0:
                self.pidOfQemu = self.getpid(pidfile.name)
                self.mtda.debug(2, "power.qemu.start(): "
                                   "qemu process started "
                                   "[{0}]".format(self.pidOfQemu))
                return True
            else:
                self.mtda.debug(1, "power.qemu.start(): "
                                   "qemu process failed "
                                   "({0})".format(result))
        return False

    def kill(self, name, pid, wait_before_kill=True, timeout=3):
        tries = timeout
        while wait_before_kill and tries > 0 and psutil.pid_exists(pid):
            self.mtda.debug(2, "waiting {0} more seconds for {1} "
                               "[{2}] to terminate".format(timeout, name, pid))
            time.sleep(1)
            tries = tries - 1
        if psutil.pid_exists(pid):
            self.mtda.debug(2, "terminating {0} "
                               "[{1}] using SIGTERM".format(name, pid))
            os.kill(pid, signal.SIGTERM)
            tries = timeout
        while tries > 0 and psutil.pid_exists(pid):
            time.sleep(1)
            tries = tries - 1
        if psutil.pid_exists(pid):
            self.mtda.debug(2, "terminating {0} "
                               "[{1}] using SIGKILL".format(name, pid))
            os.kill(pid, signal.SIGKILL)
        return psutil.pid_exists(pid)

    def stop(self):
        self.mtda.debug(3, "power.qemu.stop()")

        self.lock.acquire()
        result = True

        if self.pidOfQemu is not None:

            with open("/tmp/qemu-mtda.in", "w") as f:
                f.write("quit\n")
            result = self.kill("qemu", self.pidOfQemu)
            if result:
                self.pidOfQemu = None

        if self.pidOfSwTpm is not None:
            result = self.kill("swtpm", self.pidOfSwTpm, False)
            if result:
                self.pidOfSwTpm = None

        if self.pidOfWebsockify is not None:
            result = self.kill("websockify", self.pidOfWebsockify, False)
            if result:
                self.pidOfWebsockify = None

        self.lock.release()
        return result

    def monitor_output_non_blocking(self):
        self.mtda.debug(4, "power.qemu.monitor_output_non_blocking()")

        fd = os.open("/tmp/qemu-mtda.out", os.O_RDONLY)
        os.set_blocking(fd, False)
        try:
            output = os.read(fd, 2048).decode('utf-8')
        except BlockingIOError:
            output = ""
        os.close(fd)
        return output

    def monitor_command_output(self):
        self.mtda.debug(3, "power.qemu.monitor_command_output()")

        output = ""
        while output.endswith("(qemu) ") is False:
            output += self.monitor_output_non_blocking()
        if output.endswith("(qemu) "):
            output = output[:-7]

        self.mtda.debug(3, "power.qemu.monitor_command_output(): %s" % output)
        return output

    def _cmd(self, what):
        self.mtda.debug(3, "power.qemu._cmd()")

        started = self.start()
        if started is False:
            return None

        # flush monitor output
        self.monitor_output_non_blocking()

        # send requested command to "out" pipe
        what += "\n"
        with open("/tmp/qemu-mtda.in", "w") as f:
            f.write(what)

        # provide response from the monitor
        output = self.monitor_command_output()

        self.mtda.debug(3, "power.qemu._cmd(): %s" % str(output))
        return output

    def cmd(self, what):
        self.mtda.debug(3, "power.qemu.cmd()")

        self.lock.acquire()
        result = self._cmd(what)
        self.lock.release()

        self.mtda.debug(3, "power.qemu.cmd(): %s" % str(result))
        return result

    def command(self, args):
        self.mtda.debug(3, "power.qemu.command()")

        result = self.cmd(" ".join(args))
        result = "\n".join(result.splitlines()[1:])

        self.mtda.debug(3, "power.qemu.command(): %s" % str(result))
        return result

    def on(self):
        self.mtda.debug(3, "power.qemu.on()")

        s = self.status()
        if s == self.POWER_ON:
            return True
        self.cmd("system_reset")
        self.cmd("cont")
        return self.status() == self.POWER_ON

    def off(self):
        self.mtda.debug(3, "power.qemu.off()")

        s = self.status()
        if s == self.POWER_OFF:
            return True
        self.cmd("stop")
        self.cmd("system_reset")
        return self.status() == self.POWER_OFF

    def status(self):
        self.mtda.debug(3, "power.qemu.status()")

        result = self.POWER_UNSURE
        status = self.cmd('info status')
        if status is not None:
            for line in status.splitlines():
                line = line.strip()
                if line.startswith("VM status:"):
                    if 'running' in line:
                        result = self.POWER_ON
                    elif 'paused' in line:
                        result = self.POWER_OFF
                    break

        if result == self.POWER_UNSURE:
            self.mtda.debug(1, "unknown power status: %s" % str(status))

        self.mtda.debug(3, "power.qemu.status(): %s" % str(result))
        return result

    def usb_ids(self):
        info = self._cmd("info usb")
        lines = info.splitlines()
        results = []
        for line in lines:
            line = line.strip()
            if line.startswith("Device "):
                self.mtda.debug(2, "power.qemu.usb_ids(): {0}".format(line))
                match = re.findall(r'ID: (\S+)$', line)
                if match:
                    results.append(match[0])
        return results

    def usb_add(self, id, file):
        self.mtda.debug(3, "power.qemu.usb_add()")

        result = None
        self.lock.acquire()

        if id not in self.usb_ids():
            self.mtda.debug(2, "power.qemu.usb_add(): "
                               "adding '{0}' as '{1}'".format(file, id))
            cmdstr = "drive_add 0 if=none,id={0},file={1}"
            output = self._cmd(cmdstr.format(id, file))
            added = False
            reason = "drive_add failed"
            for line in output.splitlines():
                line = line.strip()
                if line == "OK":
                    added = True
                    break
            if added is True:
                reason = "device_add failed"
                cmdstr = "device_add usb-storage,id={0},drive={0},removable=on"
                self._cmd(cmdstr.format(id))
                added = (id in self.usb_ids())
            if added is True:
                result = id
                self.mtda.debug(2, "power.qemu.usb_add(): "
                                   "usb-storage '{0}' connected".format(id))
            else:
                self.mtda.debug(1, "power.qemu.usb_add(): "
                                   "usb-storage '{0}' could not be added "
                                   "({1})!".format(id, reason))

        self.mtda.debug(3, "power.qemu.usb_add(): %s" % str(result))
        self.lock.release()
        return result

    def usb_rm(self, id):
        self.mtda.debug(3, "power.qemu.usb_rm()")

        result = True
        self.lock.acquire()

        if id in self.usb_ids():
            self._cmd("device_del {0}".format(id))
            result = (id not in self.usb_ids())
            if result:
                self.mtda.debug(2, "power.qemu.usb_rm(): "
                                   "usb-storage '{0}' removed".format(id))
            else:
                self.mtda.debug(1, "power.qemu.usb_rm(): "
                                   "usb-storage '{0}' could not be "
                                   "removed!".format(id))

        self.mtda.debug(3, "power.qemu.usb_rm(): %s" % str(result))
        self.lock.release()
        return result


def instantiate(mtda):
    return QemuController(mtda)



# ---------------------------------------------------------------------------
# GPIO power driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import gpiod
from operator import itemgetter

# Local imports
from mtda.power.controller import PowerController


class GpioPowerController(PowerController):

    def __init__(self, mtda):
        self.dev = None
        self.mtda = mtda
        self.lines = []
        self.gpiopair = []

    def configure(self, conf):
        self.mtda.debug(3, "power.gpio.configure()")

        result = None

        if 'gpio' in conf:
            for gpio in conf['gpio'].split(','):
                self.gpiopair.append(itemgetter(0, 1)(gpio.split('@')))

        self.mtda.debug(3, "power.gpio.configure(): {}".format(result))
        return result

    def probe(self):
        self.mtda.debug(3, "power.gpio.probe()")

        result = False

        if not self.gpiopair:
            raise ValueError("GPIO chip(s) and pin(s) pair not configured")

        for chip, pin in self.gpiopair:
            pin = int(pin)
            chip = gpiod.Chip(chip, gpiod.Chip.OPEN_BY_NAME)
            self.mtda.debug(3, "this is chip {} and pin is {} "
                               "power.gpio.configure(): ".format(chip, pin))
            self.lines.append(chip.get_line(pin))

        for line in self.lines:
            try:
                if line.is_used() is False:
                    self.mtda.debug(3, "power.gpiochip{}@pin{} is free for "
                                       "use" .format(chip, pin))
                    line.request(consumer='mtda', type=gpiod.LINE_REQ_DIR_OUT)
                else:
                    raise ValueError("gpiochip{}@pin{} is in use by other "
                                     "service" .format(chip, pin))
            except OSError:
                self.mtda.debug(3, "line {} is not configured correctly"
                                   .format(line))

        result = True

        self.mtda.debug(3, "power.gpio.probe(): {}".format(result))
        return result

    def command(self, args):
        self.mtda.debug(3, "power.gpio.command()")

        result = False

        self.mtda.debug(3, "power.gpio.command(): {}".format(result))
        return result

    def on(self):
        self.mtda.debug(3, "power.gpio.on()")

        result = False

        for line in self.lines:
            line.set_value(1)
        result = self.status() == self.POWER_ON

        self.mtda.debug(3, "power.gpio.on(): {}".format(result))
        return result

    def off(self):
        self.mtda.debug(3, "power.gpio.off()")

        result = False

        for line in self.lines:
            line.set_value(0)
        result = self.status() == self.POWER_OFF

        self.mtda.debug(3, "power.gpio.off(): {}".format(result))
        return result

    def status(self):
        self.mtda.debug(3, "power.gpio.status()")

        first = True
        result = self.POWER_UNSURE

        for line in self.lines:
            value = line.get_value()
            if value == 1:
                value = self.POWER_ON
            else:
                value = self.POWER_OFF
            self.mtda.debug(3, "power.gpio.status: "
                               "line {} is {}" .format(line, value))
            if first is True:
                first = False
                result = value
            elif value != result:
                result = self.POWER_UNSURE

        self.mtda.debug(3, "power.gpio.status(): {}".format(result))
        return result


def instantiate(mtda):
    return GpioPowerController(mtda)



# ---------------------------------------------------------------------------
# Helper classes for writing pytest units
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import os
import pytest
import re
import sys
import threading
import time

import mtda.client
import mtda.constants as CONSTS
from mtda.console.screen import ScreenOutput

pytest.mtda = None


class Test:
    def init():
        if pytest.mtda is not None:
            return True

        # Initialize client
        pytest.logging = False
        pytest.mtda = mtda.client.Client()
        pytest.output = TestOutput()
        remote = pytest.mtda.remote()
        pytest.mtda.console_remote(remote, pytest.output)
        pytest.mtda.monitor_remote(remote, pytest.output)
        pytest.mtda.start()

        # Start test-suite with the DUT powered off
        assert Target.off() is True

        # Initialize boot media
        image = os.getenv("TEST_IMAGE", Consts.DEFAULT_IMAGE)
        if os.path.exists(image) is False:
            return False

        Storage.to_host()
        Storage.write(image)
        Storage.to_target()
        Env.set("boot-from-usb", "1")

        return True

    def initialized():
        return pytest.mtda is not None

    def setup():
        assert Test.init() is True

        Console.clear()
        Console.unmute()
        Config.set_power_timeout(CONSTS.DEFAULTS.POWER_TIMEOUT)
        Config.set_session_timeout(CONSTS.DEFAULTS.SESSION_TIMEOUT)

    def teardown():
        assert Test.initialized() is True

        Console.mute()
        Console.clear()


class Config:
    def set_power_timeout(timeout):
        return pytest.mtda.config_set_power_timeout(timeout)

    def set_session_timeout(timeout):
        return pytest.mtda.config_set_session_timeout(timeout)


class Consts:
    # Default image
    DEFAULT_IMAGE = "test-image.wic.img"

    # Timeouts (in seconds)
    POWER_TIMEOUT = 30


class Console:
    def clear():
        return pytest.mtda.console_clear()

    def dump():
        return pytest.mtda.console_dump()

    def flush():
        return pytest.mtda.console_flush()

    def head():
        return pytest.mtda.console_head()

    def lines():
        return pytest.mtda.console_lines()

    def mute():
        pytest.logging = False

    def send(what):
        return pytest.mtda.console_send(what)

    def tail():
        return pytest.mtda.console_tail()

    def unmute():
        pytest.logging = True

    def wait_for(what, errors=None, timeout=30, intervals=5,
                 screen="", flush=True):
        if isinstance(what, list) is False:
            what = [what]

        contents = screen
        result = None

        err_list = []
        if errors:
            if isinstance(errors, list) is False:
                errors = [errors]
            for expr in errors:
                err_list.append(re.compile(expr))

        exp_list = []
        for expr in what:
            exp_list.append(re.compile(expr))

        while timeout > 0:
            time.sleep(intervals)
            if flush is True:
                contents += pytest.mtda.console_flush()
            else:
                contents = screen + pytest.mtda.console_dump()
            # Check for errors
            for expr in err_list:
                if expr.search(contents):
                    break
            # Check for expected messages
            for expr in exp_list:
                if expr.search(contents):
                    result = contents
                    break
            if result is not None:
                break
            timeout -= intervals
        return result


class Env:
    def set(name, value):
        return pytest.mtda.env_set(name, value)


class Storage:
    def to_host():
        return pytest.mtda.storage_to_host()

    def to_target():
        return pytest.mtda.storage_to_target()

    def write(image):
        return pytest.mtda.storage_write_image(image)


class TestOutput(ScreenOutput):
    def __init__(self):
        super().__init__(pytest.mtda)
        self._power_event = threading.Event()

    def on_event(self, event):
        # Print events to stderr
        print(event, file=sys.stderr)

        # Parse event data
        info = event.split()
        domain = info[0]

        # Check for POWER events
        if domain == "POWER" and self._power_event is not None:
            self._power_event.set()

    def clear_power(self):
        self._power_event.clear()

    def wait_power(self):
        result = self._power_event.wait(Consts.POWER_TIMEOUT)
        self._power_event.clear()
        return result

    def write(self, data):
        if pytest.logging is True:
            sys.stdout.buffer.write(data)
            sys.stdout.buffer.flush()


class Target:
    def off():
        pytest.output.clear_power()
        old_status = pytest.mtda.target_status()
        result = pytest.mtda.target_off()
        if result is True and old_status != "OFF":
            result = pytest.output.wait_power()
        return result

    def on():
        pytest.output.clear_power()
        old_status = pytest.mtda.target_status()
        result = pytest.mtda.target_on()
        if result is True and old_status != "ON":
            result = pytest.output.wait_power()
        return result

    def status():
        return pytest.mtda.target_status()

    def uptime():
        return pytest.mtda.target_uptime()



#!/usr/bin/python
# ---------------------------------------------------------------------------
# Setup script for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

from setuptools import setup, find_packages

from mtda import __version__

setup(
    name='mtda',
    version=__version__,
    scripts=['mtda-cli', 'mtda-systemd-helper', 'mtda-ui', 'mtda-config'],
    packages=find_packages(exclude=["demos"]),
    package_data={'mtda': ['assets/*.*', 'templates/*.html']},
    author='Cedric Hombourger',
    author_email='cedric.hombourger@siemens.com',

    maintainer='Cedric Hombourger',
    maintainer_email='cedric.hombourger@siemens.com',

    description='Multi-Tenant Device Access',
    long_description='''
mtda is a small agent abstracting hardware controls and interfaces for a
connected test device. The main purpose of this tool is to allow developers
and testers to remotely access and control hardware devices.
''',
    url='https://github.com/siemens/mtda',
    license='MIT',
    keywords='remote test',
    classifiers=[
        "Topic :: Utilities",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.0",
        "Topic :: Software Development :: Embedded Systems",
    ],

    include_package_data=True,

    install_requires=[
        "docker",
        "flask_socketio",
        "gevent",
        "gevent-websocket",
        "kconfiglib",
        "pyserial>=2.6",
        "python-daemon>=2.0",
        "pyusb>=1.0",
        "pyzmq>=15.0,<23.0.0",
        "psutil",
        "requests",
        "urwid",
        "zerorpc>=0.6.0",
        "zstandard>=0.14",
        "zeroconf>=0.28.6"
    ],
)



# ---------------------------------------------------------------------------
# Power interface for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import abc
import mtda.constants as CONSTS


class PowerController(object):
    __metaclass__ = abc.ABCMeta

    POWER_OFF = CONSTS.POWER.OFF
    POWER_ON = CONSTS.POWER.ON
    POWER_UNSURE = CONSTS.POWER.UNSURE
    POWER_LOCKED = CONSTS.POWER.LOCKED

    @abc.abstractmethod
    def configure(self, conf):
        """ Configure this power controller from the provided configuration"""
        return

    @abc.abstractmethod
    def probe(self):
        """ Check presence of the power controller"""
        return

    @abc.abstractmethod
    def command(self, args):
        """ Send a command to the device"""
        return False

    @abc.abstractmethod
    def on(self):
        """ Power on the attached device"""
        return

    @abc.abstractmethod
    def off(self):
        """ Power off the attached device"""
        return

    @abc.abstractmethod
    def status(self):
        """ Determine the current power state of the attached device"""
        return self.POWER_OFF

    @abc.abstractmethod
    def wait(self):
        """ Wait for the target to be powered on"""
        return



# ---------------------------------------------------------------------------
# Console interface for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2021 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import abc


class ConsoleInterface(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def configure(self, conf, role='console'):
        """ Configure this console from the provided configuration"""
        return False

    @abc.abstractmethod
    def probe(self):
        """ Check presence of the console"""
        return False

    @abc.abstractmethod
    def open(self):
        """ Open the console interface"""
        return False

    @abc.abstractmethod
    def close(self):
        """ Close the console interface"""
        return

    @abc.abstractmethod
    def pending(self):
        """ Return number of pending bytes to read"""
        return 0

    @abc.abstractmethod
    def read(self, n=1):
        """ Read from the console"""
        return 0

    @abc.abstractmethod
    def write(self, data):
        """ Write to the console"""
        return 0



# ---------------------------------------------------------------------------
# Helper class for images
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import atexit
import os
import pathlib
import psutil
import subprocess
import threading

# Local imports
import mtda.constants as CONSTS
from mtda.storage.controller import StorageController


class Image(StorageController):

    def __init__(self, mtda):
        self.mtda = mtda
        self.handle = None
        self.isfuse = False
        self.isloop = False
        self.lock = threading.Lock()
        atexit.register(self._umount)

    def _close(self):
        self.mtda.debug(3, "storage.helpers.image._close()")

        result = True
        if self.handle is not None:
            self.handle.close()
            self.handle = None
            try:
                subprocess.check_output(["sync"])
            except subprocess.CalledProcessError:
                result = False

        self.mtda.debug(3, "storage.helpers.image._close(): %s" % str(result))
        return result

    def close(self):
        self.mtda.debug(3, "storage.helpers.image.close()")
        self.lock.acquire()

        result = self._close()

        self.mtda.debug(3, "storage.helpers.image.close(): %s" % str(result))
        self.lock.release()
        return result

    def _mountpoint(self, path=""):
        result = "/media"
        if os.geteuid() != 0:
            result = os.path.join("/run", "user", str(os.getuid()))
        result = os.path.join(result, "mtda", "storage")
        if path:
            result = os.path.join(result, os.path.basename(path))
        return result

    def _umount(self):
        self.mtda.debug(3, "storage.helpers.image._umount()")

        result = True
        if self._status() == CONSTS.STORAGE.ON_HOST:
            basedir = self._mountpoint()
            if os.path.exists(basedir):
                mounts = [
                    d for d in os.listdir(basedir)
                    if os.path.ismount(os.path.join(basedir, d))]
                mounts.sort()
                mounts.reverse()
                for m in mounts:
                    self.mtda.debug(2, "storage.helpers.image.umount(): "
                                       "removing mount point for '{0}'"
                                       .format(m))
                    m = os.path.join(basedir, m)
                    cmd = ["/bin/umount", m]
                    if os.geteuid() != 0:
                        cmd.insert(0, "sudo")
                    if os.system(" ".join(cmd)) == 0:
                        os.rmdir(m)
            if self.isfuse:
                device = self.device[:-1]
                self.mtda.debug(2, "storage.helpers.image.umount(): "
                                   "removing FUSE mount '{0}'".format(device))
                cmd = ["/bin/umount", device]
                if os.system(" ".join(cmd)) == 0:
                    os.rmdir(device)
                    self.device = None
                    self.isfuse = False
            elif self.isloop:
                self.mtda.debug(2, "storage.helpers.image.umount(): "
                                   "removing loopback device '{0}'"
                                   .format(self.device))
                cmd = ["losetup", "-d", self.device]
                if os.geteuid() != 0:
                    cmd.insert(0, "sudo")
                if os.system(" ".join(cmd)) == 0:
                    self.device = None
                    self.isloop = False

        self.mtda.debug(3, "storage.helpers.image._umount(): %s" % str(result))
        return result

    def umount(self):
        self.mtda.debug(3, "storage.helpers.image.umount()")
        self.lock.acquire()

        result = self._umount()

        self.mtda.debug(3, "storage.helpers.image.umount(): %s" % str(result))
        self.lock.release()
        return result

    def _get_partitions(self):
        self.mtda.debug(3, "storage.helpers.image._get_partitions()")

        self.device = None
        self.isfuse = False
        self.isloop = False
        p = pathlib.Path(self.file)
        if self.mtda.fuse is True and os.path.exists("/usr/bin/partitionfs"):
            device = os.path.join(
                "/run", "user", str(os.getuid()), "mtda", "storage", "0")
            os.makedirs(device, exist_ok=True)
            cmd = "/usr/bin/partitionfs -s {0} {1}".format(self.file, device)
            self.mtda.debug(2, "storage.helpers.image._get_partitions(): "
                               "{0}".format(cmd))
            result = (os.system(cmd)) == 0
            if result:
                self.device = device + "/"
                self.isfuse = True
        elif p.is_block_device() is True:
            self.device = self.file
            result = True
        else:
            cmd = ["losetup", "-f", "--show", "-P", self.file]
            if os.geteuid() != 0:
                cmd.insert(0, "sudo")
            device = subprocess.check_output(cmd).decode("utf-8").strip()
            result = device != ""
            if result:
                self.device = device
                self.isloop = True

        self.mtda.debug(3, "storage.helpers.image._get_partitions(): "
                           "%s" % str(result))
        return result

    def _mount_part(self, path):
        self.mtda.debug(3, "storage.helpers.image._mount_part()")

        cmd = None
        mountpoint = self._mountpoint(path)
        result = False

        if os.path.exists(path) is False:
            self.mtda.debug(1, "storage.helpers.image._mount_part(): "
                               "{0} does not exist!".format(path))
        elif os.path.ismount(mountpoint) is False:
            os.makedirs(mountpoint, exist_ok=True)
            if pathlib.Path(path).is_block_device():
                cmd = ["/bin/mount", path, mountpoint]
                if os.geteuid() != 0:
                    cmd.insert(0, "sudo")
            elif self.isfuse:
                cmd = None
                fstype = subprocess.check_output(["/usr/bin/file", path])
                fstype = fstype.decode("utf-8").strip()
                if 'ext4 filesystem' in fstype:
                    cmd = ["/usr/bin/fusext2", path, mountpoint, "-o", "rw+"]
                elif 'FAT (32 bit)' in fstype:
                    cmd = ["/usr/bin/fusefat", path, mountpoint]
                else:
                    self.mtda.debug(1, "storage.helpers.image._mount_part(): "
                                       "{0}".format(fstype))
                    self.mtda.debug(1, "storage.helpers.image._mount_part(): "
                                       "file-system not supported")
            if cmd:
                cmd = " ".join(cmd)
                self.mtda.debug(2, "storage.helpers.image._mount_part(): "
                                   "mounting {0} on {1}"
                                   .format(path, mountpoint))
                self.mtda.debug(2, "storage.helpers.image._mount_part(): "
                                   "{0}".format(cmd))
                result = (os.system(cmd)) == 0

            if result is False:
                os.rmdir(mountpoint)
        else:
            self.mtda.debug(1, "storage.helpers.image._mount_part(): "
                               "{0} is a mount point".format(mountpoint))

        self.mtda.debug(3, "storage.helpers.image._mount_part(): "
                           "%s" % str(result))
        return result

    def _part_dev(self, path, part):
        if path[-1:].isdigit():
            return "{0}p{1}".format(path, part)
        else:
            return "{0}{1}".format(path, part)

    def mount(self, part=None):
        self.mtda.debug(3, "storage.helpers.image.mount()")
        self.lock.acquire()

        result = True
        if self._status() == CONSTS.STORAGE.ON_HOST:
            result = self._get_partitions()
            if result:
                self.mtda.debug(2, "storage.helpers.image.mount(): "
                                   "'{0}' holds partitions"
                                   .format(self.device))
            else:
                self.mtda.debug(1, "storage.helpers.image.mount(): "
                                   "failed to get partitions for '{0}'"
                                   .format(self.file))
            if result:
                path = None
                if part:
                    path = self._part_dev(self.device, part)
                else:
                    path = self.device
                if path:
                    result = self._mount_part(path)
        else:
            self.mtda.debug(1, "storage.helpers.image.mount(): "
                               "storage attached to target!")
            result = False

        self.mtda.debug(3, "storage.helpers.image.mount(): %s" % str(result))
        self.lock.release()
        return result

    def open(self):
        self.mtda.debug(3, "storage.helpers.image.open()")
        self.lock.acquire()

        result = True
        if self._status() == CONSTS.STORAGE.ON_HOST:
            if self.handle is None:
                try:
                    self.handle = open(self.file, "r+b")
                    self.handle.seek(0, 0)
                except FileNotFoundError:
                    result = False

        self.mtda.debug(3, "storage.helpers.image.open(): %s" % str(result))
        self.lock.release()
        return result

    def status(self):
        self.mtda.debug(3, "storage.helpers.image.status()")
        self.lock.acquire()

        result = self._status()

        self.mtda.debug(3, "storage.helpers.image.status(): %s" % str(result))
        self.lock.release()
        return result

    def supports_hotplug(self):
        return False

    def _locate(self, dst):
        self.mtda.debug(3, "storage.helpers.image._locate()")

        result = None
        mountpoint = self._mountpoint(self.device)
        partitions = psutil.disk_partitions()
        for p in partitions:
            if p.mountpoint.startswith(mountpoint):
                path = os.path.join(p.mountpoint, dst)
                if os.path.exists(path):
                    result = path
                    break

        self.mtda.debug(3, "storage.helpers.image._locate(): %s" % str(result))
        return result

    def update(self, dst, offset):
        self.mtda.debug(3, "storage.helpers.image.update()")

        with self.lock:
            result = False
            if self.handle is None:
                path = self._locate(dst)
                if path is not None:
                    mode = "ab" if offset > 0 else "wb"
                    self.handle = open(path, mode)
                    self.handle.seek(offset)
                    result = True
                else:
                    self.mtda.debug(1, "storage.helpers.image.update(): "
                                       "%s could not be found!" % str(dst))
                    raise FileNotFoundError(dst + " could not be found!")
            else:
                self.mtda.debug(1, "storage.helpers.image.update(): "
                                   "shared storage already opened!")

            self.mtda.debug(3, "storage.helpers.image.update(): "
                               "%s" % str(result))
            return result

    def write(self, data):
        self.mtda.debug(3, "storage.helpers.image.write()")
        self.lock.acquire()

        result = None
        if self.handle is not None:
            result = self.handle.write(data)

        self.mtda.debug(3, "storage.helpers.image.write(): %s" % str(result))
        self.lock.release()
        return result



# ---------------------------------------------------------------------------
# pduclient power driver for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

# System imports
import os

# Local imports
from mtda.power.controller import PowerController


class PduClientPowerController(PowerController):

    def __init__(self, mtda):
        self.dev = None
        self.daemon = None
        self.hostname = None
        self.mtda = mtda
        self.port = None
        self.state = self.POWER_UNSURE

    def configure(self, conf):
        """ Configure this power controller from the provided configuration"""
        if 'daemon' in conf:
            self.daemon = conf['daemon']
        if 'hostname' in conf:
            self.hostname = conf['hostname']
        if 'port' in conf:
            self.port = conf['port']

    def probe(self):
        if self.daemon is None:
            raise ValueError("machine running lavapdu-listen "
                             "not specified ('daemon' not set)!")
        if self.hostname is None:
            raise ValueError("pdu not specified ('hostname' not set)!")
        if self.port is None:
            raise ValueError("port not specified!")

    def cmd(self, what):
        client = "/usr/bin/pduclient"
        return os.system(
            "{0} --daemon {1} --hostname {2} --command {3} "
            "--port {4}"
            .format(client, self.daemon, self.hostname, what, self.port))

    def command(self, args):
        return False

    def on(self):
        """ Power on the attached device"""
        status = self.cmd('on')
        if status == 0:
            self.state = self.POWER_ON
            return True
        return False

    def off(self):
        """ Power off the attached device"""
        status = self.cmd('off')
        if status == 0:
            self.state = self.POWER_OFF
            return True
        return False

    def status(self):
        """ Determine the current power state of the attached device"""
        return self.state


def instantiate(mtda):
    return PduClientPowerController(mtda)



# ---------------------------------------------------------------------------
# Scripts for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------

import importlib
from gevent import sleep


def load_device_scripts(variant, env):
    try:
        scripts = importlib.import_module("mtda.scripts." + variant)
        for op in ops.keys():
            name = "{}_{}".format(variant, op.replace('-', '_'))
            if hasattr(scripts, name) is False:
                continue
            method = getattr(scripts, name)
            if method is not None:
                ops[op][variant] = method
        for e in env.keys():
            setattr(scripts, e, env[e])
        return True
    except ImportError:
        return False


def op_handler(name):
    if name in ops and variant in ops[name]:
        mtda.debug(2, "calling '{}' device script".format(name))
        mtda.env_set(name, '0')
        result = ops[name][variant]()
        mtda.env_set(name, '{}'.format(result))
    else:
        if variant != 'unknown':
            mtda.debug(1, "no '{}' script provided "
                          "for '{}'".format(name, variant))
        mtda.env_set(name, 'not supported')


def check_op_handler(name):
    if name in env and env[name] == '1':
        op_handler(name)
        return True
    else:
        return False


def check_reset_tpm():
    return check_op_handler('reset-tpm')


def check_reset_tpm_factory():
    return check_op_handler('reset-tpm-factory')


def check_disable_secureboot():
    return check_op_handler('disable-secureboot')


def check_enable_secureboot():
    return check_op_handler('enable-secureboot')


def check_boot_from_usb():
    return check_op_handler('boot-from-usb')


def power_on():
    return op_handler('power-on')


def power_off():
    return op_handler('power-off')

# ---------------------------------------------------------------------------
# Dictionary of supported operations
# ---------------------------------------------------------------------------


ops = {
    'boot-from-usb': {},
    'disable-secureboot': {},
    'enable-secureboot': {},
    'power-on': {},
    'reset-tpm': {},
    'reset-tpm-factory': {}
}



# ---------------------------------------------------------------------------
# HomeKit support for MTDA
# ---------------------------------------------------------------------------
#
# This software is a part of MTDA.
# Copyright (C) 2023 Siemens Digital Industries Software
#
# ---------------------------------------------------------------------------
# SPDX-License-Identifier: MIT
# ---------------------------------------------------------------------------


# System imports
import os

# HAP-Python imports
from pyhap.accessory import Accessory
from pyhap.accessory_driver import AccessoryDriver
import pyhap.const as Category

# Local imports
from mtda.assistant.assistant import Assistant


class PowerSwitch(Accessory):
    category = Category.CATEGORY_OUTLET

    def __init__(self, mtda, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.mtda = mtda

        serv_switch = self.add_preload_service('Outlet')
        self.relay_on = serv_switch.configure_char(
            'On', setter_callback=self.set_relay)

        self.relay_in_use = serv_switch.configure_char(
            'OutletInUse', setter_callback=self.get_relay_in_use)

    def get_relay(self, status=None):
        if status is None:
            status = self.mtda.target_status()
        return "1" if status == "ON" else 0

    def relay_changed(self, status):
        self.mtda.debug(3, "mtda.assistant.homekit.relay_changed(%s)" % status)

        result = self.get_relay(status)
        self.relay_on.set_value(result)

        self.mtda.debug(3, "mtda.assistant.homekit.relay_changed(): "
                           "%s" % str(result))
        return result

    def set_relay(self, state):
        self.mtda.debug(3, "mtda.assistant.homekit.set_relay()")

        result = self.get_relay()
        if result != state:
            if state == 1:
                self.mtda.target_on('homekit')
            else:
                self.mtda.target_off('homekit')
            result = self.get_relay()

        self.mtda.debug(3, "mtda.assistant.homekit.set_relay(): "
                           "%s" % str(result))
        return result

    def get_relay_in_use(self, state):
        return True

    def setup_message(self):
        self.mtda.debug(3, "mtda.assistant.homekit.setup_message()")

        pincode = self.driver.state.pincode.decode()
        result = self.mtda.env_set('homekit-setup-code', pincode, 'homekit')

        self.mtda.debug(3, "mtda.assistant.homekit.setup_message(): "
                           "%s" % str(result))
        return result


class HomeKitAssistant(Assistant):

    def __init__(self, mtda):
        self.mtda = mtda
        self.accessory = None
        self.name = "MTDA"
        self.port = 51826
        self.state = "/var/lib/mtda/homekit.state"

    def configure(self, conf):
        self.mtda.debug(3, "mtda.assistant.homekit.configure()")

        result = True
        if 'name' in conf:
            self.name = conf['name']
        if 'port' in conf:
            self.port = int(conf['port'], 10)
        if 'state' in conf:
            self.state = conf['state']
        dir = os.path.dirname(self.state)
        os.makedirs(dir, mode=0o755, exist_ok=True)

        self.mtda.debug(3, "mtda.assistant.homekit.configure(): "
                           "%s" % str(result))
        return result

    def probe(self):
        return True

    def power_changed(self, status):
        self.accessory.relay_changed(status)

    def start(self):
        drv = AccessoryDriver(persist_file=self.state, port=self.port)
        self.accessory = PowerSwitch(self.mtda, drv, self.name)
        drv.add_accessory(self.accessory)
        drv.start_service()


def instantiate(mtda):
    return HomeKitAssistant(mtda)
